// Copyright 2016 Orus.io. All rights reserved.

package fullstack

import (
	"fmt"
	"testing"
	"time"

	_ "github.com/mattn/go-sqlite3"

	"github.com/gogo/protobuf/proto"
	"github.com/m4rw3r/uuid"
	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

func forwardWorker(ft *Fullstack, name string) (*xbus.Client, *xbus.Actor, *xbus.Subscription) {
	client, worker := ft.CreateWorker(name, []string{"forward"})
	sub, err := worker.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			r, err := apc.ReadEnvelope("default")
			if err != nil {
				ft.T.Error(err)
				return err
			}
			env, err := r.Complete(time.Second)
			ft.Logger.Noticef("Complete envelope: %s", env)
			if err != nil {
				ft.Logger.Errorf("Error in forwardWorker: %s", err)
			}
			apc.Log.EnvNoticef(apc.Request.Context, env, "Forwarding the envelope")
			err = apc.ForwardEnvelope(api.DefaultOutput, env)
			if err != nil {
				ft.Logger.Errorf("Error in forwardWorker: %s", err)
			}
			return nil
		})
	if err != nil {
		panic(err)
	}
	return client, worker, sub
}

func TestOneWorker(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)
	var hitContext *api.ProcessingContext

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	csub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			r, err := apc.ReadEnvelope("default")
			if err != nil {
				ft.T.Error(err)
				return err
			}
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return err
			}
			hitContext = apc.Request.Context
			hit <- e
			return nil
		})
	if err != nil {
		t.Fatal(err)
	}
	defer csub.Unsubscribe()

	_, worker, wsub := forwardWorker(ft, "fw")
	defer wsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	ft.InstallGraph(fmt.Sprintf(`
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: relay
  type: worker
  actorids: [%s]
  inputs: [default]
  outputs: [default]
- id: target
  type: consumer
  roles: [anyconsumer]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> relay
- relay.default -> target.default
`, worker.ID))

	sub, err := ft.Client.GetConn().SubscribeSync("xbus.core.events.process")
	assert.Nil(t, err)

	status, err := emitter.SendItems(
		nil, api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	// Wait for a process to change status to 'done' (ours)
	for true {
		msg, err := sub.NextMsg(time.Second)
		assert.Nil(t, err)
		if err != nil {
			t.Fatal("No process done")
		}

		var msgDec api.CoreEvent
		err = proto.Unmarshal(msg.Data, &msgDec)
		assert.Nil(t, err)
		if msgDec.Process.Status == api.Process_DONE {
			break
		}
	}

	var lastEnvelope *api.Envelope

	timeout := time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope = <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)

		time.Sleep(time.Second)
		pState := ft.Server.ProcessStorage().GetState(hitContext.ProcessID)

		for node, state := range pState.NodeStates {
			fmt.Println(" -", node, ": ")
			fmt.Println("  ", state)
		}

		assert.Equal(t, api.Process_DONE, pState.Status, "Invalid process state: %#v", pState)
	}

	// Make sure the log message landed correctly
	var logEntry model.LogEntry
	q := ft.DB.Query(ft.Model.LogEntry)
	q = q.Where(ft.Model.LogEntry.EnvelopeID.Eq(uuid.UUID(lastEnvelope.ID)))
	assert.Nil(t, q.One(&logEntry))
	assert.Equal(t, lastEnvelope.ID, api.UUID(logEntry.EnvelopeID))
}

func routerWorker(ft *Fullstack, name string) (*xbus.Client, *xbus.Actor, *xbus.Subscription) {
	c, worker := ft.CreateWorker(name, []string{"router"})
	sub, err := worker.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			r, err := apc.ReadEnvelope(api.DefaultInput)
			if err != nil {
				ft.Logger.Errorf("%s", err)
				return err
			}
			env, err := r.Complete(time.Second)
			if err != nil {
				ft.Logger.Errorf("Error in routerWorker: %s", err)
				return err
			}
			output := ""
			closeOutput := ""
			if string(env.Events[0].Items[0]) == "\"item1\"" {
				output = "output_1"
				closeOutput = "output_2"
			} else {
				output = "output_2"
				closeOutput = "output_1"
			}
			if err := apc.ForwardEnvelope(output, env); err != nil {
				ft.Logger.Errorf("Error in routerWorker: %s", err)
				return err
			}
			if err := apc.CloseOutput(closeOutput); err != nil {
				ft.Logger.Errorf("Error in routerWorker: %s", err)
				return err
			}
			// Will close automatically output_3
			return nil
		},
	)
	if err != nil {
		panic(err)
	}
	return c, worker, sub
}

func TestRouting(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit1 := make(chan *api.Envelope, 1)
	hit2 := make(chan *api.Envelope, 1)

	_, consumer1 := ft.CreateConsumer("jojo", []string{"consumerkind1"})
	sub1, err := consumer1.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			r, err := apc.ReadEnvelope(api.DefaultInput)
			if err != nil {
				ft.Logger.Errorf("%s", err)
				return err
			}
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return err
			}
			hit1 <- e
			return nil
		})
	if err != nil {
		t.Fatal(err)
	}

	defer sub1.Unsubscribe()

	_, consumer2 := ft.CreateConsumer("bugs", []string{"consumerkind2"})
	sub2, err := consumer2.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			r, err := apc.ReadEnvelope(api.DefaultInput)
			if err != nil {
				ft.Logger.Errorf("%s", err)
				return err
			}
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return err
			}
			hit2 <- e
			return nil
		})
	if err != nil {
		t.Fatal(err)
	}
	defer sub2.Unsubscribe()

	_, consumer3 := ft.CreateConsumer("last", []string{"consumerkind3"})
	sub3, err := consumer3.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			r, err := apc.ReadEnvelope(api.DefaultInput)
			if err != nil {
				ft.Logger.Errorf("%s", err)
				return err
			}
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return err
			}
			hit2 <- e
			return nil
		})
	if err != nil {
		t.Fatal(err)
	}
	defer sub3.Unsubscribe()

	_, _, subw := routerWorker(ft, "router-1")
	defer subw.Unsubscribe()

	ft.InstallGraph(fmt.Sprintf(`
name: routing
version: 0.1
status: active
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: router
  type: worker
  actors: [router-1]
  inputs: [default]
  outputs: [output_1, output_2, output_3]
- id: target1
  type: consumer
  roles: [consumerkind1]
  rolebroadcast: true
  inputs: [default]
- id: target2
  type: consumer
  roles: [consumerkind2]
  rolebroadcast: true
  inputs: [default]
- id: target3
  type: consumer
  roles: [consumerkind3]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> router
- router.output_1 -> target1.default
- router.output_2 -> target2.default
- router.output_3 -> target3.default
`))

	pStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	status, err := emitter.SendItems(
		nil, api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	select {
	case <-time.After(time.Second):
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit1:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}

	var pStatus api.Process_Status
	for pStatus != api.Process_DONE {
		select {
		case <-time.After(2 * time.Second):
			t.Fatalf("Process never ended")
		case <-hit2:
			t.Fatalf("No envelope should have reached target_2/3")
		case eState := <-pStateChan:
			t.Logf("New envelope status: %s", eState.Status)
			pStatus = eState.Status
		}
	}
}
