package http

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"net/url"
	"time"

	"github.com/gogo/protobuf/proto"
	"github.com/nats-rpc/gogo-nrpc"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

const (
	// WebHookServiceName is the name of the webhook service in the registry
	WebHookServiceName = "http.webhook"
)

// WebHookHandler ...
type WebHookHandler struct {
	url         string
	contentType string
}

// NewWebHookHandler ...
func NewWebHookHandler(actor *xbus.Actor) xbus.ActorService {
	url := actor.Options.GetConfigEntry("http.webhook.url")
	if url == "" {
		actor.Log.Errorf("No webhook on this actor")
		return nil
	}
	contentType := actor.Options.GetConfigEntry("http.webhook.content-type")
	if contentType == "" || contentType == "json" {
		contentType = "application/json"
	} else if contentType == "proto" || contentType == "protobuf" {
		contentType = "application/x-protobuf"
	}
	handler := WebHookHandler{url: url, contentType: contentType}
	return &handler
}

// Process a request
func (wh *WebHookHandler) Process(apc *xbus.ActorProcessingContext) error {
	// POST the request to the URL
	var payload []byte
	switch wh.contentType {
	case "application/json":
		// TODO support item encodings (with 'application/json;base64')
		bytes, err := json.Marshal(apc.Request)
		if err != nil {
			return fmt.Errorf("[http.webhook] Error encoding the payload: %s", err)
		}
		payload = bytes
	case "application/x-protobuf":
		bytes, err := proto.Marshal(apc.Request)
		if err != nil {
			return fmt.Errorf("[http.webhook] Error encoding the payload %s", err)
		}
		payload = bytes
	default:
		return fmt.Errorf("[http.webhook] Unsupported content-type: %s", wh.contentType)
	}

	req, err := http.NewRequest("POST", wh.url, bytes.NewBuffer(payload))
	req = req.WithContext(apc.Context)
	req.Header.Set("Content-Type", wh.contentType)
	req.Header.Set("Accept", wh.contentType)

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		urlErr := err.(*url.Error)
		return &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] %s", urlErr.Error()),
		}
	}
	defer resp.Body.Close()
	var state api.ActorProcessingState
	if resp.StatusCode == http.StatusOK {
		contentType := resp.Header.Get("Content-Type")
		switch contentType {
		case "application/json":
			dec := json.NewDecoder(resp.Body)
			if err := dec.Decode(&state); err != nil {
				return fmt.Errorf("[http.webhook] Error decoding json response: %s", err)
			}
		case "application/x-protobuf":
			b, err := ioutil.ReadAll(resp.Body)
			if err != nil {
				return fmt.Errorf("[http.webhook] Error reading response: %s", err)
			}
			if err := proto.Unmarshal(b, &state); err != nil {
				return fmt.Errorf("[http.webhook] Error decoding protobuf response: %s", err)
			}
		case "":
			return fmt.Errorf("[http.webhook] Missing Content-Type header in the response from the webhook")
		default:
			return fmt.Errorf("[http.webhook] Invalid Content-Type header in the response from the webhook. Expected 'application/json' or 'application/x-protobuf', got %s", contentType)
		}
	} else {
		state.Status = api.ActorProcessingState_ERROR
		var (
			message = resp.Status
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message = string(buf[:n])
		} else {
			message = "Could not read body: " + err.Error()
		}
		message = fmt.Sprintf(`[http.webhook] Invalid http status. Expected 200, got %s.
The response body is:
%s`,
			resp.Status,
			message)
		state.Messages = []api.LogMessage{
			{
				Time:  time.Now(),
				Level: api.LogLevel_ERROR,
				Text:  message,
			},
		}
	}
	if state.Status == api.ActorProcessingState_PENDING {
		apc.Detach()
	} else {
		apc.End(state.Status, state.Messages)
	}
	return nil
}

func init() {
	xbus.RegisterActorService(WebHookServiceName, NewWebHookHandler)
}
