package director

import (
	"testing"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/testutils/mocks"
)

func getUUID() api.UUID {
	uuid, err := api.UUIDV4()
	if err != nil {
		panic(err)
	}
	return uuid
}

func makejob(actorID api.UUID) *Job {
	return &Job{
		Job: storage.Job{
			JobID: storage.JobID{
				ProcessID: getUUID(),
				Node: storage.ProcessNode{
					ActorID: actorID,
					NodeID:  "default",
				},
			},
			JobState: storage.JobState{
				Status: api.ActorProcessingState_PENDING,
			},
		},
	}
}

func eqJob(j1, j2 storage.Job) bool {
	if j1.JobID != j2.JobID ||
		j1.Detached != j2.Detached ||
		j1.Status != j2.Status ||
		len(j1.EnvelopeIDs) != len(j2.EnvelopeIDs) {
		return false
	}
	for i, id := range j1.EnvelopeIDs {
		if id != j2.EnvelopeIDs[i] {
			return false
		}
	}
	return true
}

func TestJobQueue(t *testing.T) {
	a1 := api.MustUUIDFromString("90ecf03e-c0ee-11e8-a568-87c5719ab838")
	a2 := api.MustUUIDFromString("a7e42438-c0ee-11e8-ba96-cb0d8a7d7f85")

	js := mocks.NewJobStorage()

	q1 := newJobQueue(js, a1, 2)
	defer q1.Close()
	q2 := newJobQueue(js, a2, 2)
	defer q2.Close()

	j11 := makejob(a1)
	j12 := makejob(a1)
	j13 := makejob(a1)
	j14 := makejob(a1)

	js.NewJob(j11.JobID, j11.JobState)
	q1.Add(j11)
	js.NewJob(j12.JobID, j12.JobState)
	q1.Add(j12)
	js.NewJob(j13.JobID, j13.JobState)
	q1.Add(j13)
	// At this point, the buffer is full and the next job will only be present
	// in storage
	js.NewJob(j14.JobID, j14.JobState)
	q1.Add(j14)

	t.Log(j11, j12, j13, j14)

	j := <-q1.C()
	if !eqJob(j.Job, j11.Job) {
		t.Errorf("Expected job1, got %v", j)
	}
	if j != j11 {
		t.Errorf("Expected the same instance (%p), got %p", j11, j)
	}
	j = <-q1.C()
	if !eqJob(j.Job, j12.Job) {
		t.Errorf("Expected %v, got %p, %v (j1 was %v)", j12, j, j, j11)
		t.Fatal()
	}
	j = <-q1.C()
	if !eqJob(j.Job, j13.Job) {
		t.Errorf("Expected %v, got %v", j13, j)
	}
	j = <-q1.C()
	if !eqJob(j.Job, j14.Job) {
		t.Errorf("Expected %v, got %v", j14, j)
	}
	if j == j14 {
		t.Errorf("Expected a new instance of j14")
	}

	q1 = newJobQueue(js, a1, 2)
	defer q1.Close()

	j = <-q1.C()
	if !eqJob(j.Job, j11.Job) {
		t.Errorf("Expected job1, got %v", j)
	}
	j = <-q1.C()
	if !eqJob(j.Job, j12.Job) {
		t.Errorf("Expected job2, got %v", j)
	}
	j = <-q1.C()
	if !eqJob(j.Job, j13.Job) {
		t.Errorf("Expected job3, got %v", j)
	}
	j = <-q1.C()
	if !eqJob(j.Job, j14.Job) {
		t.Errorf("Expected job4, got %v", j)
	}
}

func TestRingBuffer(t *testing.T) {

	t.Run("NewRingBuffer", func(t *testing.T) {
		rb := NewRingBuffer(5)
		if rb.Size != 5 {
			t.Errorf("NewRingBuffer creation failed")
		}
		if len(rb.Container) != 5 {
			t.Errorf("NewRingBuffer: wrong container size")
		}
		if rb.Empty() == false {
			t.Error("NewRingBuffer: Empty() should be true")
		}
		if rb.Full() == true {
			t.Error("NewRingBuffer: Full() should be false")
		}
	})

	t.Run("ReadingBuffer", func(t *testing.T) {
		rb := NewRingBuffer(2)
		x := rb.Read()
		if x != nil {
			t.Errorf("Empty buffer should return nil values")
		}
	})
	t.Run("WritingBuffer", func(t *testing.T) {
		j := Job{}
		rb := NewRingBuffer(2)
		rb.Write(&j)
		if rb.Empty() {
			t.Error("After a write the RingBuffer should not be empty")
		}
		if x := rb.Read(); x != &j {
			t.Errorf("Error Manipulating RingBuffer, Added %v, Got %v\n", j, x)
		}
		if !rb.Empty() {
			t.Error("RingBuffer should be empty again")
		}

		x := rb.Read()
		if x != nil {
			t.Errorf("Error Manipulating RingBuffer, No Element in Buffer, Found 1")
		}
	})

	t.Run("MultipleValues", func(t *testing.T) {
		rb := NewRingBuffer(2)
		j1 := &Job{}
		j2 := &Job{}
		j3 := &Job{}
		if rb.Count() != 0 {
			t.Errorf("Count() should be 0")
		}
		if err := rb.Write(j1); err != nil {
			t.Errorf("Write failed, got err=%s", err)
		}
		if rb.Count() != 1 {
			t.Errorf("Count() should be 1")
		}
		if err := rb.Write(j2); err != nil {
			t.Errorf("Write failed, got err=%s", err)
		}
		if rb.Count() != 2 {
			t.Errorf("Count() should be 2")
		}

		if !rb.Full() {
			t.Error("RingBuffer should be full")
		}

		if err := rb.Write(j3); err != ErrBufferFull {
			t.Errorf("Write should fail with ErrBufferFull, got err=%s", err)
		}

		a := rb.Read()
		if rb.Full() {
			t.Error("RingBuffer should not be full")
		}
		if rb.Count() != 1 {
			t.Errorf("Count() should be 1")
		}

		b := rb.Read()

		if rb.Count() != 0 {
			t.Errorf("Count() should be 0")
		}

		if a != j1 || b != j2 {
			t.Errorf("Error Manipulating RingBuffer")
		}

	})
}
