package xbus

import (
	"context"
	"errors"
	"time"

	api "bitbucket.org/orus-io/go-xbus/api"
)

var (
	// ErrNoSuchInput when invalid input name is given
	ErrNoSuchInput = errors.New("No such input")
)

// ActorProcessingContext is passed to the actor Process handlers
type ActorProcessingContext struct {
	Request      *api.ActorProcessRequest
	Context      context.Context
	send         func(api.ActorProcessingState)
	Actor        *Actor
	Log          *ActorLogger
	LocalLog     Logger
	currentState api.ActorProcessingState
	detached     bool
	done         bool
}

// ReadEnvelope returns a EnvelopeReceiver that reads the envelope of the
// given input
func (apc *ActorProcessingContext) ReadEnvelope(
	inputName string,
) (*EnvelopeReceiver, error) {
	for _, input := range apc.Request.GetInputs() {
		if input.Name == inputName {
			er := apc.Actor.Client.RetrieveEnvelopeFrom(
				apc.Context, "actor", apc.Actor.ID.String(),
				*input.Envelope, *input.Position,
			)
			return er, nil
		}
	}
	return nil, ErrNoSuchInput
}

// CompleteEnvelope returns a complete envelope.
// Equivalent of doing (with of course a better error handling):
//
//    receiver, _ = apc.ReadEnvelope(inputName)
//    envelope, _ = receiver.Complete(timeout)
func (apc *ActorProcessingContext) CompleteEnvelope(
	inputName string, timeout time.Duration,
) (*api.Envelope, error) {
	receiver, err := apc.ReadEnvelope(inputName)
	if err != nil {
		return nil, err
	}
	return receiver.Complete(timeout)
}

// OpenOutput returns a EnvelopeSender that sends to a given output
func (apc *ActorProcessingContext) OpenOutput(
	outputName string, eventTypes ...string,
) *EnvelopeSender {
	return apc.Actor.OpenOutput(
		apc.Request.Context, outputName, eventTypes...)
}

// CloseOutput closes an output without sending anything to it
func (apc *ActorProcessingContext) CloseOutput(outputName string) error {
	_, err := apc.Actor.CloseOutput(apc.Request.Context, outputName)
	return err
}

// ForwardEnvelope sends an existing envelope to an output
func (apc *ActorProcessingContext) ForwardEnvelope(
	outputName string, envelope *api.Envelope,
) error {
	newEnv := api.Envelope{
		ID:   envelope.ID,
		Last: true,
	}
	_, err := apc.Actor.SendEnvelope(
		apc.Request.Context,
		outputName,
		newEnv)
	return err
}

// Detach signals that when the handler finishes, the processing should
// not be automatically marked done successfully
func (apc *ActorProcessingContext) Detach() {
	apc.detached = true
}

// Done informs the bus that the processing is over
func (apc *ActorProcessingContext) Done() {
	// TODO handle errors
	apc.currentState.Status = api.ActorProcessingState_SUCCESS
	apc.done = true
	if apc.detached {
		apc.Actor.ProcessingSuccess(apc.Request.Context)
	} else {
		apc.send(apc.currentState)
	}
}

// ActorLeaving informs the bus that the actor will be offline just after
// handling the request
func (apc *ActorProcessingContext) ActorLeaving() {
	apc.currentState.ActorLeaving = true
}

// Done informs the bus that the processing ends with error
func (apc *ActorProcessingContext) Error(err error) {
	apc.currentState.Status = api.ActorProcessingState_ERROR
	apc.currentState.Messages = []api.LogMessage{{
		Time:  time.Now(),
		Level: api.LogLevel_ERROR,
		Text:  err.Error(),
	}}
	if apc.detached {
		apc.Actor.ProcessingError(apc.Request.Context, "Error processing: %s", err)
	} else {
		apc.send(apc.currentState)
	}
}

// End sends the 'processing end' message
func (apc *ActorProcessingContext) End(
	status api.ActorProcessingState_Status, msgs []api.LogMessage,
) {
	apc.currentState.Status = status
	apc.currentState.Messages = msgs
	apc.done = true
	if apc.detached {
		apc.Actor.ProcessingEnd(apc.Request.Context, status, msgs)
	} else {
		apc.send(apc.currentState)
	}
}
